/*
 * Generic Convex Polygon Scan Conversion and Clipping
 * by Paul Heckbert
 * from "Graphics Gems", Academic Press, 1990
 */

/*
 * poly_scan.c: point-sampled scan conversion of convex polygons
 *
 * Paul Heckbert        1985, Dec 1989
 */

#include <stdio.h>
#include <math.h>
#include "polygon.h"

static scanline(int y, vertex *l, vertex *r, window *Window,
                void (*pixelproc)(), unsigned long mask);
static incrementalize_y(register double *p1, register double *p2,
                        register double *p, register double *dp,
                        int y, register unsigned long mask);
static incrementalize_x(register double *p1, register double *p2,
                        register double *p, register double *dp,
                        int x, register unsigned long mask);
static increment(register double *p, register double *dp,
                 register unsigned long mask);
/*
 * poly_scan: Scan convert a polygon, calling pixelproc at each pixel with an
 * interpolated Poly_vert structure.  Polygon can be clockwise or ccw.
 * Polygon is clipped in 2-D to win, the screen space window.
 *
 * Scan conversion is done on the basis of Poly_vert fields sx and sy.
 * These two must always be interpolated, and only they have special meaning
 * to this code; any other fields are blindly interpolated regardless of
 * their semantics.
 *
 * The pixelproc subroutine takes the arguments:
 *
 *      pixelproc(x, y, point)
 *      int x, y;
 *      Poly_vert *point;
 *
 * All the fields of point indicated by p->mask will be valid inside pixelproc
 * except sx and sy.  If they were computed, they would have values
 * sx=x+.5 and sy=y+.5, since sampling is done at pixel centers.
 */

/* p - polygon */
/* Window - 2-D screen space clipping window */
/* pixelproc - procedure called at each pixel */

void polygon_scan_convert(register polygon *p, window *Window,
                          void (*pixelproc)())
{
   register int    i, li, ri, y, ly, ry, top, rem;
   register unsigned long mask;
   double          ymin;
   vertex       l, r, dl, dr;

   if (p->n > POLYGON_MAX_VERTICES)
   {
      fprintf(stderr, "poly_scan: too many vertices: %d\n", p->n);
      return;
   }
   if (sizeof(vertex) / sizeof(double) > 32)
   {
      fprintf(stderr, "Poly_vert structure too big; must be <=32 doubles\n");
      exit(1);
   }

   ymin = HUGE_VAL;
   for (i = 0; i < p->n; i++)   /* find top vertex (y points down) */
      if (p->Vertices[i].sy < ymin)
      {
         ymin = p->Vertices[i].sy;
         top = i;
      }

   li = ri = top;               /* left and right Verticesex indices */
   rem = p->n;                  /* number of Verticesices remaining */
   y = ceil(ymin - .5);         /* current scan line */
   ly = ry = y - 1;             /* lower end of left & right edges */
   mask = p->mask & ~VERTEX_MASK(sy);     /* stop interpolating screen y */

   while (rem > 0)
   {                            /* scan in y, activating new edges on left &
                                 * right */
      /* as scan line passes over new Verticesices */

      while (ly <= y && rem > 0)
      {                         /* advance left edge? */
         rem--;
         i = li - 1;            /* step ccw down left side */
         if (i < 0)
            i = p->n - 1;
         incrementalize_y((double *) &(p->Vertices[li]),
                          (double *) &(p->Vertices[i]),
                          (double *) &l, (double *) &dl, y, mask);
         ly = floor(p->Vertices[i].sy + .5);
         li = i;
      }
      while (ry <= y && rem > 0)
      {                         /* advance right edge? */
         rem--;
         i = ri + 1;            /* step cw down right edge */
         if (i >= p->n)
            i = 0;
         incrementalize_y((double *) &(p->Vertices[ri]),
                          (double *) &(p->Vertices[i]),
                          (double *) &r, (double *) &dr, y, mask);
         ry = floor(p->Vertices[i].sy + .5);
         ri = i;
      }

      while (y < ly && y < ry)
      {                         /* do scanlines till end of l or r edge */
         if (y >= Window->LL_y && y <= Window->UR_y)
            if (l.sx <= r.sx)
               scanline(y, &l, &r, Window, pixelproc, mask);
            else
               scanline(y, &r, &l, Window, pixelproc, mask);
         y++;
         increment((double *) &l, (double *) &dl, mask);
         increment((double *) &r, (double *) &dr, mask);
      }
   }
}

/* scanline: output scanline by sampling polygon at Y=y+.5 */

static scanline(int y, vertex *l, vertex *r, window *Window,
                void (*pixelproc)(), unsigned long mask)
{
   int             x, lx, rx;
   vertex       p, dp;

   mask &= ~VERTEX_MASK(sx);      /* stop interpolating screen x */
   lx = ceil(l->sx - .5);
   if (lx < Window->LL_x)
      lx = Window->LL_x;
   rx = floor(r->sx - .5);
   if (rx > Window->UR_x)
      rx = Window->UR_x;
   if (lx > rx)
      return;
   incrementalize_x((double *) l, (double *) r, (double *) &p, (double *) &dp,
                    lx, mask);
   for (x = lx; x <= rx; x++)
   {                            /* scan in x, generating pixels */
      (*pixelproc) (x, y, &p);
      increment((double *) &p, (double *) &dp, mask);
   }
}

/*
 * incrementalize_y: put intersection of line Y=y+.5 with edge between points
 * p1 and p2 in p, put change with respect to y in dp
 */

static incrementalize_y(register double *p1, register double *p2,
                        register double *p, register double *dp,
                        int y, register unsigned long mask)
{
   double          dy, frac;

   dy = ((vertex *) p2)->sy - ((vertex *) p1)->sy;
   if (dy == 0.)
      dy = 1.;
   frac = y + .5 - ((vertex *) p1)->sy;

   for (; mask != 0; mask >>= 1, p1++, p2++, p++, dp++)
      if (mask & 1)
      {
         *dp = (*p2 - *p1) / dy;
         *p = *p1 + *dp * frac;
      }
}

/*
 * incrementalize_x: put intersection of line X=x+.5 with edge between points
 * p1 and p2 in p, put change with respect to x in dp
 */

static incrementalize_x(register double *p1, register double *p2,
                        register double *p, register double *dp,
                        int x, register unsigned long mask)
{
   double          dx, frac;

   dx = ((vertex *) p2)->sx - ((vertex *) p1)->sx;
   if (dx == 0.)
      dx = 1.;
   frac = x + .5 - ((vertex *) p1)->sx;

   for (; mask != 0; mask >>= 1, p1++, p2++, p++, dp++)
      if (mask & 1)
      {
         *dp = (*p2 - *p1) / dx;
         *p = *p1 + *dp * frac;
      }
}

static increment(register double *p, register double *dp,
                 register unsigned long mask)
{
   for (; mask != 0; mask >>= 1, p++, dp++)
      if (mask & 1)
         *p += *dp;
}
